@@ -4,28 +4,15 @@
 */
 
 /*
- * Copyright (c) 2007-2008 Roy Marples <roy@marples.name>
+ * Copyright (c) 2007-2015 The OpenRC Authors.
+ * See the Authors file at the top-level directory of this distribution and
+ * https://github.com/OpenRC/openrc/blob/master/AUTHORS
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * This file is part of OpenRC. It is subject to the license terms in
+ * the LICENSE file found in the top-level directory of this
+ * distribution and at https://github.com/OpenRC/openrc/blob/master/LICENSE
+ * This file may not be copied, modified, propagated, or distributed
+ *    except according to the terms contained in the LICENSE file.
  */
 
 const char libeinfo_copyright[] = "Copyright (c) 2007-2008 Roy Marples";
@@ -896,7 +883,7 @@ eindent(void)
 {
 	char *env = getenv("EINFO_INDENT");
 	int amount = 0;
-	char num[10];
+	char *num;
 
 	if (env) {
 		errno = 0;
@@ -907,8 +894,9 @@ eindent(void)
 	amount += INDENT_WIDTH;
 	if (amount > INDENT_MAX)
 		amount = INDENT_MAX;
-	snprintf(num, 10, "%08d", amount);
+	xasprintf(&num, "%08d", amount);
 	setenv("EINFO_INDENT", num, 1);
+	free(num);
 }
 hidden_def(eindent)
 
@@ -916,7 +904,7 @@ void eoutdent(void)
 {
 	char *env = getenv("EINFO_INDENT");
 	int amount = 0;
-	char num[10];
+	char *num = NULL;
 	int serrno = errno;
 
 	if (!env)
@@ -930,8 +918,9 @@ void eoutdent(void)
 	if (amount <= 0)
 		unsetenv("EINFO_INDENT");
 	else {
-		snprintf(num, 10, "%08d", amount);
+		xasprintf(&num, "%08d", amount);
 		setenv("EINFO_INDENT", num, 1);
+		free(num);
 	}
 	errno = serrno;
 }
