@@ -29,7 +29,7 @@
 
 const char *applet = NULL;
 const char *extraopts = NULL;
-const char *getoptstring = "aclmrsu" getoptstring_COMMON;
+const char *getoptstring = "aclmrsSu" getoptstring_COMMON;
 const struct option longopts[] = {
 	{"all",         0, NULL, 'a'},
 	{"crashed",     0, NULL, 'c'},
@@ -37,6 +37,7 @@ const struct option longopts[] = {
 	{"manual",        0, NULL, 'm'},
 	{"runlevel",    0, NULL, 'r'},
 	{"servicelist", 0, NULL, 's'},
+	{"supervised", 0, NULL, 'S'},
 	{"unused",      0, NULL, 'u'},
 	longopts_COMMON
 };
@@ -47,6 +48,7 @@ const char * const longopts_help[] = {
 	"Show manually started services",
 	"Show the name of the current runlevel",
 	"Show service list",
+	"show supervised services",
 	"Show services not assigned to any runlevel",
 	longopts_help_COMMON
 };
@@ -167,6 +169,9 @@ print_service(const char *service)
 	} else if (state & RC_SERVICE_SCHEDULED) {
 		xasprintf(&status, "scheduled");
 		color = ECOLOR_WARN;
+	} else if (state & RC_SERVICE_FAILED) {
+		xasprintf(&status, "failed");
+		color = ECOLOR_WARN;
 	} else
 		xasprintf(&status, " stopped ");
 
@@ -238,6 +243,7 @@ print_stacked_services(const char *runlevel)
 
 int main(int argc, char **argv)
 {
+	RC_SERVICE state;
     RC_STRING *s, *l, *t, *level;
 	bool show_all = false;
 	char *p, *runlevel = NULL;
@@ -292,6 +298,14 @@ int main(int argc, char **argv)
 			printf("%s\n", runlevel);
 			goto exit;
 			/* NOTREACHED */
+		case 'S':
+			services = rc_services_in_state(RC_SERVICE_STARTED);
+			TAILQ_FOREACH_SAFE(s, services, entries, t)
+				if (!rc_service_value_get(s->value, "child_pid"))
+					TAILQ_REMOVE(services, s, entries);
+			print_services(NULL, services);
+			goto exit;
+			/* NOTREACHED */
 		case 's':
 			services = rc_services_in_runlevel(NULL);
 			print_services(NULL, services);
@@ -373,11 +387,14 @@ int main(int argc, char **argv)
 			free(nservices);
 		}
 		TAILQ_FOREACH_SAFE(s, services, entries, t) {
+			state = rc_service_state(s->value);
 			if ((rc_stringlist_find(sservices, s->value) ||
-			    (rc_service_state(s->value) & ( RC_SERVICE_STOPPED | RC_SERVICE_HOTPLUGGED)))) {
-				TAILQ_REMOVE(services, s, entries);
-				free(s->value);
-				free(s);
+			    (state & ( RC_SERVICE_STOPPED | RC_SERVICE_HOTPLUGGED)))) {
+				if (! (state & RC_SERVICE_FAILED)) {
+					TAILQ_REMOVE(services, s, entries);
+					free(s->value);
+					free(s);
+				}
 			}
 		}
 		needsme = rc_stringlist_new();
