@@ -78,9 +78,6 @@ const char *usagestring = ""					\
 #define INITSH                  RC_LIBEXECDIR "/sh/init.sh"
 #define INITEARLYSH             RC_LIBEXECDIR "/sh/init-early.sh"
 
-#define SHUTDOWN                "/sbin/shutdown"
-#define SULOGIN                 "/sbin/sulogin"
-
 #define INTERACTIVE             RC_SVCDIR "/interactive"
 
 #define DEVBOOT			"/dev/.rcboot"
@@ -104,7 +101,6 @@ clean_failed(void)
 {
 	DIR *dp;
 	struct dirent *d;
-	size_t l;
 	char *path;
 
 	/* Clean the failed services state dir now */
@@ -115,16 +111,11 @@ clean_failed(void)
 				(d->d_name[1] == '.' && d->d_name[2] == '\0')))
 				continue;
 
-			l = strlen(RC_SVCDIR "/failed/") +
-			    strlen(d->d_name) + 1;
-			path = xmalloc(sizeof(char) * l);
-			snprintf(path, l, RC_SVCDIR "/failed/%s", d->d_name);
-			if (path) {
-				if (unlink(path))
-					eerror("%s: unlink `%s': %s",
-					    applet, path, strerror(errno));
-				free(path);
-			}
+			xasprintf(&path, RC_SVCDIR "/failed/%s", d->d_name);
+			if (unlink(path))
+				eerror("%s: unlink `%s': %s",
+				    applet, path, strerror(errno));
+			free(path);
 		}
 		closedir(dp);
 	}
@@ -288,8 +279,8 @@ open_shell(void)
 	/* VSERVER systems cannot really drop to shells */
 	if (sys && strcmp(sys, RC_SYS_VSERVER) == 0)
 	{
-		execl("/sbin/halt", "/sbin/halt", "-f", (char *) NULL);
-		eerrorx("%s: unable to exec `/sbin/halt': %s",
+		execlp("halt", "halt", "-f", (char *) NULL);
+		eerrorx("%s: unable to exec `halt -f': %s",
 		    applet, strerror(errno));
 	}
 #endif
@@ -336,26 +327,26 @@ set_krunlevel(const char *level)
 	return true;
 }
 
-static size_t
-get_krunlevel(char *buffer, int buffer_len)
+static char *get_krunlevel(void)
 {
+	char *buffer = NULL;
 	FILE *fp;
 	size_t i = 0;
 
 	if (!exists(RC_KRUNLEVEL))
-		return 0;
+		return NULL;
 	if (!(fp = fopen(RC_KRUNLEVEL, "r"))) {
 		eerror("fopen `%s': %s", RC_KRUNLEVEL, strerror(errno));
-		return 0;
+		return NULL;
 	}
 
-	if (fgets(buffer, buffer_len, fp)) {
+	if (getline(&buffer, &i, fp) != -1) {
 		i = strlen(buffer);
 		if (buffer[i - 1] == '\n')
 			buffer[i - 1] = 0;
 	}
 	fclose(fp);
-	return i;
+	return buffer;
 }
 
 static void
@@ -394,7 +385,7 @@ static void
 handle_signal(int sig)
 {
 	int serrno = errno;
-	char signame[10] = { '\0' };
+	char *signame = NULL;
 	pid_t pid;
 	RC_PID *pi;
 	int status = 0;
@@ -425,16 +416,16 @@ handle_signal(int sig)
 		break;
 
 	case SIGINT:
-		if (!signame[0])
-			snprintf(signame, sizeof(signame), "SIGINT");
+		if (!signame)
+			xasprintf(&signame, "SIGINT");
 		/* FALLTHROUGH */
 	case SIGTERM:
-		if (!signame[0])
-			snprintf(signame, sizeof(signame), "SIGTERM");
+		if (!signame)
+			xasprintf(&signame, "SIGTERM");
 		/* FALLTHROUGH */
 	case SIGQUIT:
-		if (!signame[0])
-			snprintf(signame, sizeof(signame), "SIGQUIT");
+		if (!signame)
+			xasprintf(&signame, "SIGQUIT");
 		eerrorx("%s: caught %s, aborting", applet, signame);
 		/* NOTREACHED */
 	case SIGUSR1:
@@ -505,6 +496,9 @@ do_sysinit()
 	 * sys */
 	if ((sys = rc_sys()))
 		setenv("RC_SYS", sys, 1);
+	/* force an update of the dependency tree */
+	if ((main_deptree = _rc_deptree_load(1, NULL)) == NULL)
+		eerrorx("failed to load deptree");
 }
 
 static bool
@@ -512,14 +506,11 @@ runlevel_config(const char *service, const char *level)
 {
 	char *init = rc_service_resolve(service);
 	char *conf, *dir;
-	size_t l;
 	bool retval;
 
 	dir = dirname(init);
 	dir = dirname(init);
-	l = strlen(dir) + strlen(level) + strlen(service) + 10;
-	conf = xmalloc(sizeof(char) * l);
-	snprintf(conf, l, "%s/conf.d/%s.%s", dir, service, level);
+	xasprintf(&conf, "%s/conf.d/%s.%s", dir, service, level);
 	retval = exists(conf);
 	free(conf);
 	free(init);
@@ -683,6 +674,8 @@ do_start_services(const RC_STRINGLIST *start_services, bool parallel)
 		}
 
 		pid = service_start(service->value);
+		if (pid == -1)
+			break;
 		/* Remember the pid if we're running in parallel */
 		if (pid > 0) {
 			add_pid(pid);
@@ -741,8 +734,8 @@ int main(int argc, char **argv)
 	RC_STRING *service;
 	bool going_down = false;
 	int depoptions = RC_DEP_STRICT | RC_DEP_TRACE;
-	char krunlevel [PATH_MAX];
-	char pidstr[10];
+	char *krunlevel = NULL;
+	char *pidstr = NULL;
 	int opt;
 	bool parallel;
 	int regen = 0;
@@ -842,8 +835,9 @@ int main(int argc, char **argv)
 	setenv("EINFO_LOG", "openrc", 1);
 
 	/* Export our PID */
-	snprintf(pidstr, sizeof(pidstr), "%d", getpid());
+	xasprintf(&pidstr, "%d", getpid());
 	setenv("RC_PID", pidstr, 1);
+	free(pidstr);
 
 	/* Create a list of all services which should be started for the new or
 	* current runlevel including those in boot, sysinit and hotplugged
@@ -890,7 +884,8 @@ int main(int argc, char **argv)
 		    (strcmp(newlevel, RC_LEVEL_SYSINIT) != 0 &&
 			strcmp(newlevel, getenv("RC_BOOTLEVEL")) != 0))
 		{
-			if (get_krunlevel(krunlevel, sizeof(krunlevel))) {
+			krunlevel = get_krunlevel();
+			if (krunlevel) {
 				newlevel = krunlevel;
 				set_krunlevel(NULL);
 			}
